searchNodes=[{"doc":"OLEDVirtual is a virtual version of the OLED library. Its main purpose it to reduce the roundtrip time during development by fully virtualizing the oled display. See OLEDVirtual.Display for the virtual oled display. See OLEDVirtual.MultiDisplay to invoke several displays simultaneously.","ref":"OLEDVirtual.html","title":"OLEDVirtual","type":"module"},{"doc":"Virtual oled display. When used, the display expects an :app as option. The :app should be the app that has the configuration. Example defmodule MyApp.MyDisplay do use OLEDVirtual.Display , app : :my_app end Could be configured with: config :my_app , MyApp.MyDisplay , width : 128 , height : 64 It needs to be added to the supervision tree: defmodule MyApp.Application do use Application @impl true def start ( _type , _args ) do children = [ ... MyApp.MyDisplay ] opts = [ strategy : :one_for_one , name : MyApp.Supervisor ] Supervisor . start_link ( children , opts ) end end And then used like this: MyApp.MyDisplay . rect ( 0 , 0 , 127 , 63 ) MyApp.MyDisplay . display ( ) See OLED.Display for all draw and display functions. Callbacks The virtual display module supports optional callbacks to get notified about buffer or frame changes. They can be used to further process the frame, e.g. broadcasting it via Phoenix.PubSub . defmodule MyApp.MyDisplay do use OLEDVirtual.Display , app : :my_app def on_display ( data , dimensions ) do # Called after MyApp.MyDisplay.display/0 and MyApp.MyDisplay.display_frame/2 end def on_buffer_update ( data , dimensions ) do # Called after any changes to the buffer of the next frame end end","ref":"OLEDVirtual.Display.html","title":"OLEDVirtual.Display","type":"behaviour"},{"doc":"Called after any changes to the buffer of the next frame","ref":"OLEDVirtual.Display.html#c:on_buffer_update/2","title":"OLEDVirtual.Display.on_buffer_update/2","type":"callback"},{"doc":"Called after display/0 and display_frame/2 got invoked.","ref":"OLEDVirtual.Display.html#c:on_display/2","title":"OLEDVirtual.Display.on_display/2","type":"callback"},{"doc":"A set of helper functions to format the frame data.","ref":"OLEDVirtual.Format.html","title":"OLEDVirtual.Format","type":"module"},{"doc":"Formats the frame data into a list of bit values. A value of 1 means the pixel is on and a value of 0 means the pixel is off. { :ok , frame } = MyApp.MyDisplay . get_frame ( ) bits = OLEDVirtual.Format . as_bits ( frame )","ref":"OLEDVirtual.Format.html#as_bits/1","title":"OLEDVirtual.Format.as_bits/1","type":"function"},{"doc":"Formats the frame data into a nested array with rows and columns. A value of 1 means the pixel is on and a value of 0 means the pixel is off. { :ok , width , _height } = MyApp.MyDisplay . get_dimensions ( ) { :ok , frame } = MyApp.MyDisplay . get_frame ( ) pixel_matrix = OLEDVirtual.Format . as_matrix ( frame , width )","ref":"OLEDVirtual.Format.html#as_matrix/2","title":"OLEDVirtual.Format.as_matrix/2","type":"function"},{"doc":"Multi display to invoke several displays simultaneously. It supports the same functions as OLEDVirtual.Display and OLED.Display . The only difference is that the functions always return a list of {display, result} tuples, where display is the display module like MyApp.MyDisplay and result is the function result. The display functions are invoked simultaneously using Task.async/1 , so they do not block each other. When used, the multi display expects an :app as option. The :app should be the app that has the configuration. Example defmodule MyApp.MyMultiDisplay do use OLEDVirtual.MultiDisplay , app : :my_app end Could be configured with: config :my_app , MyApp.MyMultiDisplay , displays : [ MyApp.OledVirtual , MyAppFirmware.Oled , ] And then used like this: MyApp.MyMultiDisplay . rect ( 0 , 0 , 127 , 63 ) MyApp.MyMultiDisplay . display ( ) See OLED.Display for all draw and display functions. Telemetry Events Each function call on each display emits a telemetry event. [:oled_virtual, :multi_display, &lt;function_name&gt;] Where &lt;function_name&gt; is the invoked function name as an atom, e.g. :display The event contains the following measurements: :duration - The duration of the function call in milliseconds The event contains the following metadata: :display - The display module :multi_display - The multi display module","ref":"OLEDVirtual.MultiDisplay.html","title":"OLEDVirtual.MultiDisplay","type":"module"},{"doc":"","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"Initial release","ref":"changelog.html#1-0-0","title":"Changelog - 1.0.0","type":"extras"},{"doc":"The entire point of this library is to provide a virtual abstraction of the hardware OLED display. But for proper usage, the virtual OLED display actually needs to be displayed somewhere, preferably via a web interface. This guide will show an example using Phoenix LiveView as it is probably the simplest solution when using a Phoenix project anyway.","ref":"virtual-display-liveview.html","title":"Virtual Display Using LiveView","type":"extras"},{"doc":"The idea is simple. The virtual display module supports an on_display/2 callback, which is called whenever a new frame needs to be rendered. The callback can be used to broadcast the frame data using Phoenix.PubSub , which a LiveView can subscribe to. The LiveView then pushes the frame data via an event to the browser client. A custom LiveView hook on the client receives the event and renders the frame on a canvas.","ref":"virtual-display-liveview.html#high-level-concept","title":"Virtual Display Using LiveView - High Level Concept","type":"extras"},{"doc":"The on_display/2 callback on the display module gets called on each new frame, which will be used to broadcast the frame data and dimensions using Phoenix.PubSub . # my_app/lib/my_app/oled_virtual_display.ex defmodule MyApp.OledVirtualDisplay do use OLEDVirtual.Display , app : :my_app def on_display ( data , dimensions ) do payload = %{ data : data , dimensions : dimensions } Phoenix.PubSub . broadcast ( MyApp.PubSub , &quot;oled-virtual&quot; , %{ event : &quot;on_display&quot; , payload : payload } ) end end","ref":"virtual-display-liveview.html#broadcasting-new-frames","title":"Virtual Display Using LiveView - Broadcasting New Frames","type":"extras"},{"doc":"Below is the entire LiveView module. It sends an initial setup event to the browser client to render the first frame, then listens for new frames to also send them to the browser client. # my_app/lib/my_app_web/live/oled_virtual.ex defmodule MyAppWeb.OledVirtualLive do use MyAppWeb , :live_view alias MyAppWeb.Endpoint alias MyApp.OledVirtualDisplay alias OLEDVirtual.Format @impl true def render ( assigns ) do ~H&quot;&quot;&quot; &lt;div id={&quot;oled-virtual-\#{@id}&quot;} phx-hook=&quot;OledVirtual&quot;&gt; &lt;div phx-update=&quot;ignore&quot;&gt; &lt;div data-element=&quot;canvas-container&quot;&gt;&lt;/div&gt; &lt;/div&gt; &lt;/div&gt; &quot;&quot;&quot; end @impl true def mount ( _params , _session , socket ) do if connected? ( socket ) do Endpoint . subscribe ( &quot;oled-virtual&quot; ) end { :ok , width , height } = OledVirtualDisplay . get_dimensions ( ) { :ok , frame } = OledVirtualDisplay . get_frame ( ) setup_event_payload = %{ data : Format . as_bits ( frame ) , dimensions : %{ width : width , height : height } } socket = socket |&gt; assign ( width : width ) |&gt; assign ( height : height ) |&gt; assign ( id : socket . id ) |&gt; push_event ( &quot;setup&quot; , setup_event_payload ) { :ok , socket , layout : { MyAppWeb.LayoutView , &quot;live_empty.html&quot; } } end @impl true def handle_info ( %{ event : &quot;on_display&quot; , payload : %{ data : data } } , socket ) do payload = %{ data : Format . as_bits ( data ) } { :noreply , push_event ( socket , &quot;new_frame&quot; , payload ) } end end The mount/3 function sets a custom live_empty.html layout, which lives next to the default live.html . Its usage is optional, but makes for a cleaner HTML. # my_app/lib/my_app_web/templates/layout/live_empty.html.heex &lt; % = @inner_content % &gt;","ref":"virtual-display-liveview.html#adding-the-liveview","title":"Virtual Display Using LiveView - Adding the LiveView","type":"extras"},{"doc":"The template in the LiveView module above uses a custom OledVirtual hook on the outer div . Below is the entire LiveView hook. It handles the canvas creation in the setup event handler and updates the canvas on every new_frame event. // my_app/assets/js/app.js let Hooks = {} Hooks.OledVirtual = { mounted() { const container = this.el.querySelector('[data-element=&quot;canvas-container&quot;]'); this.handleEvent(&quot;setup&quot;, (payload) =&gt; { const existingCanvas = container.getElementsByTagName('canvas') ; if (existingCanvas.length) { Array.from(existingCanvas).map((c) =&gt; c.remove()); } const canvas = document.createElement('canvas'); canvas.width = payload.dimensions.width; canvas.height = payload.dimensions.height; const scale = 2; canvas.style.imageRendering = 'pixelated'; canvas.style.transformOrigin = 'top left'; canvas.style.transform = `scale(${scale})`; container.style.width = `${payload.dimensions.width * scale}px`; container.style.height = `${payload.dimensions.height * scale}px`; const ctx = canvas.getContext('2d'); container.appendChild(canvas); displayImage(ctx, payload.dimensions, payload.data); this.ctx = ctx; this.dimensions = payload.dimensions; }); this.handleEvent(&quot;new_frame&quot;, (payload) =&gt; { displayImage(this.ctx, this.dimensions, payload.data); }); } } const displayImage = (ctx, dimensions, data) =&gt; { const imgData = ctx.createImageData(dimensions.width, dimensions.height); let i; for (i = 0; i &lt; imgData.data.length; i += 1) { const color = data[i] === 1 ? 255 : 0; imgData.data[(i * 4)] = color; imgData.data[1 + (i * 4)] = color; imgData.data[2 + (i * 4)] = color; imgData.data[3 + (i * 4)] = 255; } ctx.putImageData(imgData, 0, 0); } let liveSocket = new LiveSocket(&quot;/live&quot;, Socket, { hooks: Hooks, params: {_csrf_token: csrfToken}, })","ref":"virtual-display-liveview.html#adding-the-liveview-hook","title":"Virtual Display Using LiveView - Adding the LiveView Hook","type":"extras"},{"doc":"The LiveView can be rendered inside or outside existing live views using the live_render function. &lt;%= live_render ( @conn , MyAppWeb.OledVirtualLive , id : &quot;oled-virtual&quot; ) %&gt; The exact place depends on the specific application. A simple setup might have it directly inside root.html.heex so it is always visible on the web-interface.","ref":"virtual-display-liveview.html#using-the-liveview","title":"Virtual Display Using LiveView - Using the LiveView","type":"extras"},{"doc":"Probably the most common use case for the MultiDisplay module is in a Nerves project where it invokes the virtual display during local development and both virtual and hardware display on the device. This way, a potential web interface could always provide a preview of the image on the display. To build the actual virtual display using Phoenix LiveView , check out this guide .","ref":"multidisplay-and-nerves.html","title":"Multi Display and Nerves","type":"extras"},{"doc":"For this guide, we use a Nerves typical poncho structure. my_app | -- firmware ` -- my_app The inner my_app folder contains a normal Phoenix project and the firmware folder the Nerves project. For a more in-depth guide on this setup, refer to the official Nerves + Phoenix guide . The additional dependencies for both projects are as follows: # my_app/firmware/mix.exs defp deps do [ ... { :oled , &quot;~&gt; 0.3.5&quot; } , { :circuits_i2c , &quot;~&gt; 1.0&quot; , override : true } ] end # my_app/my_app/mix.exs defp deps do [ ... { :oled_virtual , &quot;~&gt; 1.0&quot; } ] end","ref":"multidisplay-and-nerves.html#project-setup","title":"Multi Display and Nerves - Project Setup","type":"extras"},{"doc":"The OLED display will be created inside the firmware project because it needs to interact with the real hardware which is not available during local development. # my_app/firmware/lib/my_app_firmware/oled_display.ex defmodule MyAppFirmware.OledDisplay do use OLED.Display , app : :my_app_firmware end # my_app/firmware/config/target.exs config :my_app_firmware , MyAppFirmware.OledDisplay , device : &quot;i2c-1&quot; , driver : :ssd1306 , type : :i2c , width : 128 , height : 64 , rst_pin : 25 , dc_pin : 24 , address : 0x3C This configuration is specific to the OLED Bonnet display. Your OLED display might need a slightly different configuration. Please refer to the OLED documentation. The MyAppFirmware.OledDisplay module needs to be added to the supervision tree. # my_app/firmware/lib/my_app_firmware/application.ex def children ( _target ) do [ MyAppFirmware.OledDisplay ] end","ref":"multidisplay-and-nerves.html#adding-the-oled-display","title":"Multi Display and Nerves - Adding the OLED Display","type":"extras"},{"doc":"The virtual OLED display will be part of the my_app Phoenix project. # my_app/my_app/lib/my_app/oled_virtual_display.ex defmodule MyApp.OledVirtualDisplay do use OLEDVirtual.Display , app : :my_app def on_display ( data , dimensions ) do # React to new frames end end # my_app/my_app/config/config.exs config :my_app , MyApp.OledVirtualDisplay , width : 128 , height : 64 Make sure that the configured resolution is the same as on the hardware display. The MyApp.OledVirtualDisplay module needs to be added to the supervision tree. # my_app/my_app/lib/my_app/application.ex @impl true def start ( _type , _args ) do children = [ ... MyApp.OledVirtualDisplay ] opts = [ strategy : :one_for_one , name : MyApp.Supervisor ] Supervisor . start_link ( children , opts ) end","ref":"multidisplay-and-nerves.html#adding-the-virtual-oled-display","title":"Multi Display and Nerves - Adding the Virtual OLED Display","type":"extras"},{"doc":"The multi display will also be part of the my_app Phoenix project, and invoke either the virtual display or both displays, based on the specific configuration that is loaded during runtime. # my_app/my_app/lib/my_app/multi_display.ex defmodule MyApp.MultiDisplay do use OLEDVirtual.MultiDisplay , app : :my_app end Now, we need to configure it in both projects. # my_app/my_app/config/config.exs config :my_app , MyApp.MultiDisplay , displays : [ MyApp.OledVirtualDisplay ] # my_app/firmware/config/target.exs config :my_app , MyApp.MultiDisplay , displays : [ MyApp.OledVirtualDisplay , MyAppFirmware.OledDisplay ] This is all it takes. From now on, all display interactions can be made through the MyApp.MultiDisplay module, which invokes the configured display modules. If the Nerves firmware is built using the firmware project, the multi display will invoke both displays. And if the my_app Phoenix project is started locally for development, only the virtual display will be invoked.","ref":"multidisplay-and-nerves.html#adding-the-multi-display","title":"Multi Display and Nerves - Adding the Multi Display","type":"extras"}]